;DCF77 Clock with extended display (15 digits: time, date, weekday)
;Controller: ATTiny2313 @ 4.194304 MHz external crystal
;Fuses for ATTiny2313: low=0xDD, high=0xDB, ext=0xFF
;
;Assemble with avra (http://avra.sourceforge.net/) only, this file contains
;directives specific for this assembler!
;
;(C) Arne Rossius, http://www.ebps.de.vu
;
;DISPLAY:
;  - 14 LED displays, 7 segments + dp + 7 LEDs for weekday (15th "display")
;  - common cathode, segments connected through PNP transistors
;  - common cathode lines using 2x ULN2003 (ULN2803 + ULN2003 with weekday)
;  - shift register for common lines, first output = left display
;  - shift register outouts are active high for ULN2003/ULN2803 inputs
;  - display order: HH:MM:SS YYYY-MM-DD [Weekday] (or DD-MM-YYYY, set below)
;  - weekday display (optional): 7 LEDs, output a = Monday, ..., g = Sunday
;  - second LEDs (optional): 60 LEDs, shift register, first output = LED 59
;  - weekday matrix (optional): rotated column cathode 5x7 dot matrix display
;
;PORTS:
;  PB0-6 = Segments (a..g), active-low
;  PD2   = DCF77 input
;  PD3   = "enable display" switch (open: display off, closed to GND: on)
;
;   Pin  | Regular Clock       | Second LEDs   | Weekday Matrix
;  ------+---------------------+---------------+----------------------------
;   PB7  | dp (active low)     | <===          | Matrix Column 7
;   PD0  | --                  | Data output   | Matrix Column 1
;   PD1  | --                  | Clock output  | Matrix Column 2
;   PD4  | UTC Switch (to GND) | Strobe output | Matrix Column 3
;   PD5  | Data for Shift Reg. | <===          | <=== *and* Matrix Column 5
;   PD6  | Clock for Shift Reg.| <===          | <=== *and* Matrix Column 6
;
;VERSION HISTORY:
;  v0.2  2007-01-10  Arne Rossius
;  v0.3  2007-02-04  Arne Rossius
;    * added switch to disable display
;  v0.4  2008-03-15  Arne Rossius
;    * weekday output on virtual display #15
;    * auto-increment date and weekday
;    * settings for inverted outputs and reverse date format
;  v0.5  2008-04-04
;    * display leading zero for day of the month
;  v0.6  2008-04-22  Arne Rossius
;    * BUGFIX: auto-increment date: load ZH:ZL instead of YH:ZL (narf!)
;    * added option to disable weekday display (slightly improves brightness)
;  v0.7  2008-07-25  Arne Rossius
;    * BUGFIX: added support for inverted common outputs (setting existed since
;              version 0.4, but didn't have an effect)
;  v0.8  2008-08-09  Arne Rossius
;    * BUGFIX: all segments were lit when "disable display" switch was open and
;              "InvertSegments" setting enabled
;    * changed name of "DisableWeekday" setting to "WeekdayLEDs" and reversed
;      meaning of 0 and 1 (previously 1 meant disabled, now 1 means enabled)
;    * changed method for display blanking when "disable display" switch open
;    * added support for 7x5 dot matrix weekday display (decimal points have to
;      be wired differently if this option is used)
;  v0.8.1  2008-11-13  Arne Rossius
;    * added support for AT90S1200 controller (except for weekday matrix)
;    * replaced month length table by month length calculation
;    * don't require weekday matrix files if weekday matrix not used
;  v0.8.2  2009-06-28  Arne Rossius
;    * added support for second LEDs (60 LEDs, one for each second)
;  v0.9  2010-02-08  Arne Rossius
;    * new DCF77 receiver code (with plausibility check)
;    * removed AT90S1200 support
;    * added UTC support
;  v0.9.1  2010-02-26  Arne Rossius
;    * replaced direct RAM offsets by RAM offset constants

;microcontroller definition file
.include "../include/2313def.inc"

;DCF77 receiver
.equ	DCF_INVERTED = 1 ;inverted DCF77 signal (low pulses)
.equ	DCF_PULLUP = 1 ;internal pullup for DCF77 input
.equ	DCFInput = 2 ;don't change!

;switches
.equ	SWPIN = PIND
.equ	SWDisplay = 3
.equ	SWUTC = 4

;LED display
.equ	SEGMENT = PORTB
.equ	COMMON = PORTD
.equ	DATA = 5
.equ	CLOCK = 6
.equ	SEC_LED_DATA = 0
.equ	SEC_LED_CLOCK = 1
.equ	SEC_LED_STROBE = 4
.equ	MATRIX1 = 0
.equ	MATRIX2 = 1
.equ	MATRIX3 = 4
.equ	MATRIX5 = 5
.equ	MATRIX6 = 6
.equ	MATRIX7 = 7

;GENERAL SETTINGS
.equ	ReverseDateFormat = 0 ;set to 1 for DD-MM-YYYY date format
.equ	InvertCommon = 0 ;set to 1 for active-low common outputs (for PNP)
.equ	InvertSegments = 0 ;set to 1 for active-high segment outputs (NPN)
.equ	UTCDefault = 0 ;set to 1 for UTC display (local time when PD4 = GND)
;WEEKDAY LEDS (disadvantage: very slightly darker display)
.equ	WeekdayLEDs = 1 ;set to 1 to enable virtual digit 15 for 7 weekday LEDs
;SECOND LEDS (disadvantage: no UTC switch possible)
.equ	SecondLEDs = 0 ;set to 1 to enable second LEDs
.equ	SecondLEDsMode = 1 ;0 = single LED, 1 = building up, 2 = building up and down
.equ	InvertSecondLEDs = 0 ;set to 1 for active-low second LED outputs
;WEEKDAY MATRIX (disadvantages: no decimal points and no UTC switch possible)
.equ	WeekdayMatrix = 0 ;set to 1 to enable 7x5 dot matrix weekday display
.equ	InvertMatrix = 0 ;set to 1 for active-high weekday matrix outputs (NPN)

;needed for lpm instruction
.def	lpmresult = R0

;time/date registers
.def	subsec = R1
.def	seconds = R2

;general registers
.def	temp = R16
.def	temp2 = R17

;DCF77-specific registers
.def	dcfbits = R18
.def	dcfflags = R19

;additional registers for multiplexing
.def	display = R20 ;active display [0..13]
.def	matrix = R21 ;weekday matrix data for PORTD

;R29:R28 = Y pointer (RAM)
;R31:R30 = Z pointer (Flash)

;*******************************************************************************
;*
;*   MAIN PROGRAM
;*
;*******************************************************************************

;---------------------------------------
; Interrupt vectors
;---------------------------------------

.cseg

;Reset
.org 0x000
	rjmp	start
	
;Timer 0 overflow
.org OVF0addr
	rjmp	ovf0int

.equ DCF_USE_INT0 = 1
.include "dcf77.asm"

;---------------------------------------
; Initialization
;---------------------------------------
	
start:
	;stack pointer
	ldi	temp, LOW(RAMEND)
	out	SPL, temp
	
	;disable analog comparator
	ldi	temp, 0x80
	out	ACSR, temp
	
	;I/O ports
	ldi	temp, 0xFF
	out	DDRB, temp
	out	PORTB, temp
	ldi	temp, 0x73
	out	DDRD, temp
	.if (DCF_PULLUP == 0)
	ldi	temp, 0x7B
	.else
	ldi	temp, 0x7F
	.endif
	out	PORTD, temp
	
	;timer
	ldi	temp, 1<<TOIE0 ;overflow int on
	out	TIMSK, temp
	ldi	temp, 0x04 ;Clk/256 = 64Hz int
	out	TCCR0, temp
	
	;registers
	ldi	display, 0
	
	rcall	dcf_init
	
	;enable interrupts
	sei

;---------------------------------------
; Main loop
;   modifies: temp, temp2, display, ZL, ZH, lpmresult
;---------------------------------------

;multiplex all the time and do *all* the action while one of the displays is enabled
main_loop:
	;enable next display
.if (InvertSegments == 0)	;display off
	ldi	temp2, 0xFF
.else
	ldi	temp2, 0x00
.endif
	out	SEGMENT, temp2
	cpi	display, 0	;select next display
	sbis	SWPIN, SWDisplay ;no display if switch is open
	breq	PC+3
.if ( InvertCommon == 0 )
	cbi	COMMON, DATA
	rjmp	PC+2
	sbi	COMMON, DATA
.else
	sbi	COMMON, DATA
	rjmp	PC+2
	cbi	COMMON, DATA
.endif
	cbi	COMMON, CLOCK
	sbi	COMMON, CLOCK
	out	COMMON, matrix ;display on (output new data)
	out	SEGMENT, temp
	
	;update second LEDs
.if (SecondLEDs != 0)
	sbrs	dcfflags, dcfTimeOK
	rjmp	sec_leds_end
  .if (WeekdayMatrix != 0)
	.error "Second LEDs and Weekday Matrix can't be enabled at the same time!"
  .endif
	;convert seconds to binary
	cbi	COMMON, SEC_LED_STROBE
	mov	temp, seconds
	rcall	bcd2bin
	inc	temp
	ldi	temp2, 60
sec_leds_loop:
  .if (SecondLEDsMode == 0)
	.message "Second LEDs: single LED"
    .if (InvertSecondLEDs == 0)
	cbi	COMMON, SEC_LED_DATA
	cp	temp, temp2
	brne	PC+2
	sbi	COMMON, SEC_LED_DATA
    .else
	sbi	COMMON, SEC_LED_DATA
	cp	temp, temp2
	brne	PC+2
	cbi	COMMON, SEC_LED_DATA
    .endif
  .elif (SecondLEDsMode == 1)
	.message "Second LEDs: building up"
    .if (InvertSecondLEDs == 0)
	cbi	COMMON, SEC_LED_DATA
	cp	temp, temp2
	brlo	PC+2
	sbi	COMMON, SEC_LED_DATA
    .else
	sbi	COMMON, SEC_LED_DATA
	cp	temp, temp2
	brlo	PC+2
	cbi	COMMON, SEC_LED_DATA
    .endif
  .elif (SecondLEDsMode == 2)
	.message "Second LEDs: building up and down"
    .if (InvertSecondLEDs == 0)
	sbrs	minutes, 0
	cbi	COMMON, SEC_LED_DATA
	sbrc	minutes, 0
	sbi	COMMON, SEC_LED_DATA
	cp	temp, temp2
	brlo	PC+5
	sbrs	minutes, 0
	sbi	COMMON, SEC_LED_DATA
	sbrc	minutes, 0
	cbi	COMMON, SEC_LED_DATA
    .else
	sbrs	minutes, 0
	sbi	COMMON, SEC_LED_DATA
	sbrc	minutes, 0
	cbi	COMMON, SEC_LED_DATA
	cp	temp, temp2
	brlo	PC+5
	sbrs	minutes, 0
	cbi	COMMON, SEC_LED_DATA
	sbrc	minutes, 0
	sbi	COMMON, SEC_LED_DATA
    .endif
  .else
	.error "Wrong setting for SecondLEDsMode!"
  .endif
	cbi	COMMON, SEC_LED_CLOCK
	sbi	COMMON, SEC_LED_CLOCK
	dec	temp2
	brne	sec_leds_loop
	sbi	COMMON, SEC_LED_STROBE
sec_leds_end:
.endif
	
	;encode the next display
	inc	display
.if ( WeekdayLEDs == 0 )
	cpi	display, 14 ;return to display 0 after display 13 (no weekday LEDs)
.else
	cpi	display, 15 ;return to display 0 after display 14
.endif
	brlo	PC+2
	ldi	display, 0
	
	;set data for display
.if (UTCDefault == 0)
	ldi	YL, RAM_Time
  .if ((WeekdayMatrix == 0) && (SecondLEDs == 0))
	sbis	SWPIN, SWUTC
	ldi	YL, RAM_UTC
  .endif
.else
	ldi	YL, RAM_UTC
  .if ((WeekdayMatrix == 0) && (SecondLEDs == 0))
	sbis	SWPIN, SWUTC
	ldi	YL, RAM_Time
  .endif
.endif
	cpi	display, 2
	brlo	disp_hr
	cpi	display, 4
	brlo	disp_min
	cpi	display, 6
	brlo	disp_sec
.if (ReverseDateFormat == 0)
	cpi	display, 8 ;YYYY-MM-DD format
	brlo	disp_20
	cpi	display, 10
	brlo	disp_yr
	cpi	display, 12
	brlo	disp_mon
	cpi	display, 14
	brlo	disp_day
.else
	cpi	display, 8 ;DD-MM-YYYY format
	brlo	disp_day
	cpi	display, 10
	brlo	disp_mon
	cpi	display, 12
	brlo	disp_20
	cpi	display, 14
	brlo	disp_yr
.endif
	ldd	ZL, Y+tWeekday	;weekday
.if (InvertSegments == 0)
	ldi	temp, 0xFE
	sbrs	dcfflags, dcfTimeOK
	ldi	temp, 0xFF ;" "
.else
	ldi	temp, 0x01
	sbrs	dcfflags, dcfTimeOK
	ldi	temp, 0x00 ;" "
.endif
disp_wkday_loop:
	cpi	ZL, 2
	brsh	PC+2
	rjmp	encode_end
	lsl	temp
.if (InvertSegments == 0)
	ori	temp, 0x01
.endif
	dec	ZL
	rjmp	disp_wkday_loop

disp_20:			;century (20)
	ldi	ZL, 0x20
	sbrs	dcfflags, dcfTimeOK
	ldi	ZL, 0xFF ;"  "
	rjmp	display_end
disp_yr:			;year
	ldd	ZL, Y+tYear
	sbrs	dcfflags, dcfTimeOK
	ldi	ZL, 0xFF ;"  "
	rjmp	display_end
disp_mon:			;month
	ldd	ZL, Y+tMonth
	sbrs	dcfflags, dcfTimeOK
	ldi	ZL, 0xFF ;"  "
	rjmp	display_end
disp_day:
	ldd	ZL, Y+tDay	;day (or DCF77 input)
	sbrc	dcfflags, dcfTimeOK
	rjmp	display_end
	ldi	ZL, 0xFD ;" _"
	in	temp, MCUCR
	.if (DCF_INVERTED == 0)
	sbrs	temp, ISC00
	.else
	sbrc	temp, ISC00
	.endif
	ldi	ZL, 0xFE ;" -"
	rjmp	display_end
disp_hr:			;hour (or # of bits received)
	ldd	ZL, Y+tHour
	sbrc	dcfflags, dcfTimeOK
	rjmp	display_end
	mov	ZL, dcfbits
	rcall	bin2bcd
	rjmp	display_end
disp_sec:			;second (or last received bit)
	mov	ZL, seconds
	sbrc	dcfflags, dcfTimeOK
	rjmp	display_end
	ldi	ZL, 0xFE ;" -"
	cpi	dcfbits, 0
	breq	display_end
	sbrc	dcfflags, dcfStop
	rjmp	display_end
	ldi	ZL, 0xF0 ;" 0"
	lds	temp, RAM_DCF2+5
	sbrc	temp, 7
	ldi	ZL, 0xF1 ;" 1"
	rjmp	display_end
disp_min:			;minute (or last error)
	ldd	ZL, Y+tMinute
	sbrc	dcfflags, dcfTimeOK
	rjmp	display_end
	ldi	ZL, 0xFE ;" -" => no error, receiving
	sbrc	dcfflags, dcfFirstTelegram
	ldi	ZL, 0xF1 ;" 1" => first telegram OK, receiving second one
	sbrc	dcfflags, dcfStop
	ldi	ZL, 0xF0 ;" 0" => DCF reception stopped, but no error
	mov	temp, dcfflags
	andi	temp, dcfErrMask
	cpi	temp, dcfErrBitCount
	brne	PC+2
	ldi	ZL, 0xFA ;" C" => Bit Count error
	cpi	temp, dcfErrBitLength
	brne	PC+2
	ldi	ZL, 0xFB ;" L" => Bit Length error
	cpi	temp, dcfErrParity
	brne	PC+2
	ldi	ZL, 0xF2 ;" 2" => Parity error
	cpi	temp, dcfErrStartBit
	brne	PC+2
	ldi	ZL, 0xF5 ;" S" => Start Bit error
	cpi	temp, dcfErrDST
	brne	PC+2
	ldi	ZL, 0xF3 ;" 3" => invalid DST Bit combination
	cpi	temp, dcfErrPlausibility
	brne	PC+2
	ldi	ZL, 0xFC ;" P" => Plausibility Error
	rjmp	display_end

display_end:

	;suppress leading zero
	cpi	display, 0
	brne	PC+4
	cpi	ZL, 0x10
	brsh	PC+2
	ldi	ZL, 0xF0 ;15 = space
	
	;encode to 7-segment
	sbrs	display, 0 ;swap if 10's wanted
	swap	ZL
	andi	ZL, 0x0F
	clr	ZH
	ldi	temp, LOW(charset*2)
	add	ZL, temp
	ldi	temp, HIGH(charset*2)
	adc	ZH, temp
	lpm
	mov	temp, lpmresult
	
	;enable dp after hours, minutes, year and month
	sbrs	dcfflags, dcfTimeOK
	rjmp	dp_end
	cpi	display, 5
	breq	dp_end
.if (ReverseDateFormat == 0)
	cpi	display, 7
.else
	cpi	display, 11
.endif
	breq	dp_end
	cpi	display, 13
	breq	dp_end
	sbrc	display, 0
.if (InvertSegments == 0)
	andi	temp, ~0x80
.else
	ori	temp, 0x80
.endif
dp_end:
encode_end:

	;generate weekday matrix data
.if (InvertMatrix == 0)
	.if (DCF_PULLUP == 0)
	ldi	matrix, 0xFB
	.else
	ldi	matrix, 0xFF
	.endif
.else
	.if (DCF_PULLUP == 0)
	ldi	matrix, 1<<SWDisplay
	.else
	ldi	matrix, 1<<DCFInput | 1<<SWDisplay ;keep pullups on inputs enabled
	.endif
.endif
	cpi	display, 9
	brlo	matrix_end
	cpi	display, 14
	brsh	matrix_end
	ldd	temp2, Y+tWeekday ;day of week
	mov	ZL, temp2 ;ZH:ZL = wkday * 5
	lsl	ZL
	lsl	ZL
	add	ZL, temp2
	clr	ZH
	ldi	temp2, LOW(weekdays*2-9)
	add	ZL, temp2
	ldi	temp2, HIGH(weekdays*2-9)
	adc	ZH, temp2
	ldi	temp2, 0
	add	ZL, display
	adc	ZH, temp2
	lpm
	mov	matrix, lpmresult
	.if (DCF_PULLUP == 0)
	ori	matrix, 1<<SWDisplay
	.else
	ori	matrix, 1<<DCFInput | 1<<SWDisplay ;keep pullups on inputs enabled
	.endif
matrix_end:

	;add weekday matrix data for PB7
.if (WeekdayMatrix != 0)
  .if (InvertMatrix == 0)
	ori	temp, 0x80
	sbrc	matrix, 7
	andi	temp, 0x7F
  .else
	andi	temp, 0x7F
	sbrc	matrix, 7
	ori	temp, 0x80
  .endif
.else
  .if (SecondLEDs == 0)
	ori	matrix, 1<<SWUTC ;pullup for UTC switch
  .endif
.endif
	
	;wait for a while
	ldi	ZH, 2
	ldi	ZL, 0
multiplex_wait:
	dec	ZL
	brne	multiplex_wait
	dec	ZH
	brne	multiplex_wait
	
	rjmp	main_loop
	
;---------------------------------------
; 7-Segment encoding table
;---------------------------------------

charset:
	.equ	CHARSET_0 = 0x3F ; 0
	.equ	CHARSET_1 = 0x06 ; 1
	.equ	CHARSET_2 = 0x5B ; 2
	.equ	CHARSET_3 = 0x4F ; 3
	.equ	CHARSET_4 = 0x66 ; 4
	.equ	CHARSET_5 = 0x6D ; 5
	.equ	CHARSET_6 = 0x7D ; 6
	.equ	CHARSET_7 = 0x07 ; 7
	.equ	CHARSET_8 = 0x7F ; 8
	.equ	CHARSET_9 = 0x6F ; 9
	.equ	CHARSET_A = 0x39 ; C
	.equ	CHARSET_B = 0x38 ; L
	.equ	CHARSET_C = 0x73 ; P
	.equ	CHARSET_D = 0x08 ; _
	.equ	CHARSET_E = 0x40 ; -
	.equ	CHARSET_F = 0x00 ; (Space)
.if (InvertSegments == 0)
	.db ~CHARSET_0, ~CHARSET_1, ~CHARSET_2, ~CHARSET_3
	.db ~CHARSET_4, ~CHARSET_5, ~CHARSET_6, ~CHARSET_7
	.db ~CHARSET_8, ~CHARSET_9, ~CHARSET_A, ~CHARSET_B
	.db ~CHARSET_C, ~CHARSET_D, ~CHARSET_E, ~CHARSET_F
.else
	.db CHARSET_0, CHARSET_1, CHARSET_2, CHARSET_3
	.db CHARSET_4, CHARSET_5, CHARSET_6, CHARSET_7
	.db CHARSET_8, CHARSET_9, CHARSET_A, CHARSET_B
	.db CHARSET_C, CHARSET_D, CHARSET_E, CHARSET_F
.endif

;---------------------------------------
; Weekday Matrix encoding table
;---------------------------------------

weekdays:
.if (WeekdayMatrix != 0)
  .if (InvertMatrix == 0)
	.include "weekdays/german-uc.asm"
  .else
	.include "weekdays/german-uc-inv.asm"
  .endif
.endif

;*******************************************************************************
;*
;*   DCF77 subroutines
;*
;*******************************************************************************

;---------------------------------------
; Timer 0 overflow int (64Hz)
;   modifies: subsec, seconds
;---------------------------------------

ovf0int:
	;save registers
	push	temp
	push	temp2
	push	YL
	in	temp, SREG
	push	temp
	
;---------------------------------------
; increment time (fallback quartz clock)
;---------------------------------------

	;increment 1/64-seconds
	inc	subsec
	ldi	temp, 64
	cp	subsec, temp
	brlo	ovf0_inctime_end
	
	;increment seconds
	clr	subsec
	mov	temp, seconds
	rcall	bcdinc
	mov	seconds, temp
	cpi	temp, 0x60
	brlo	ovf0_inctime_end
	clr	seconds
	
	ldi	YL, RAM_Time
	rcall	inctime
	rcall	calculate_utc
ovf0_inctime_end:
	
	pop	temp
	out	SREG, temp
	pop	YL
	pop	temp2
	pop	temp
	reti

;*******************************************************************************
;*
;*   GENERAL ARITHMETIC FUNCTIONS
;*
;*******************************************************************************

;---------------------------------------
; convert binary to packed BCD
;   in: ZL
;   out: ZL
;   modifies: temp
;---------------------------------------

bin2bcd:
	ldi	temp, 0
bin2bcd_loop:
	subi	temp, -0x10
	subi	ZL, 10
	brcc	bin2bcd_loop
	subi	temp, 0x10
	subi	ZL, -10
	add	ZL, temp
	ret
	
;---------------------------------------
; convert packed BCD to binary
;   in: temp
;   out: temp
;   modifies: temp2
;---------------------------------------

bcd2bin:
	mov	temp2, temp
	andi	temp, 0x0F
	swap	temp2
	andi	temp2, 0x0F
	lsl	temp2
	add	temp, temp2 ;temp += temp2*2
	lsl	temp2
	lsl	temp2
	add	temp, temp2 ;temp += temp2*8
	ret
